package display

const (
	PanelBrowser             = "browser"
	PanelBookmarks           = "bkmk"
	PanelDownload            = "dl"
	PanelDownloadChoiceModal = "dlChoice"
	PanelHelp                = "help"

	PanelYesNoModal = "yesno"
	PanelInfoModal  = "info"
	PanelErrorModal = "error"
	PanelInputModal = "input"
)
